summaryrefslogtreecommitdiffstats
path: root/src/core/hle/kernel
diff options
context:
space:
mode:
authorSubv <subv2112@gmail.com>2018-05-19 23:58:30 +0200
committerSubv <subv2112@gmail.com>2018-05-19 23:58:30 +0200
commit2a35a3625195087b4e7b0e420b91637e31ef3bca (patch)
treed451b368237da2be47a2122ade4fe6489ff17c0e /src/core/hle/kernel
parentKernel/Threads: Reschedule the proper core when operating on that core's threads. (diff)
downloadyuzu-2a35a3625195087b4e7b0e420b91637e31ef3bca.tar
yuzu-2a35a3625195087b4e7b0e420b91637e31ef3bca.tar.gz
yuzu-2a35a3625195087b4e7b0e420b91637e31ef3bca.tar.bz2
yuzu-2a35a3625195087b4e7b0e420b91637e31ef3bca.tar.lz
yuzu-2a35a3625195087b4e7b0e420b91637e31ef3bca.tar.xz
yuzu-2a35a3625195087b4e7b0e420b91637e31ef3bca.tar.zst
yuzu-2a35a3625195087b4e7b0e420b91637e31ef3bca.zip
Diffstat (limited to 'src/core/hle/kernel')
-rw-r--r--src/core/hle/kernel/svc.cpp119
1 files changed, 68 insertions, 51 deletions
diff --git a/src/core/hle/kernel/svc.cpp b/src/core/hle/kernel/svc.cpp
index 810ef66aa..0811a16b8 100644
--- a/src/core/hle/kernel/svc.cpp
+++ b/src/core/hle/kernel/svc.cpp
@@ -608,61 +608,78 @@ static ResultCode SignalProcessWideKey(VAddr condition_variable_addr, s32 target
NGLOG_TRACE(Kernel_SVC, "called, condition_variable_addr=0x{:X}, target=0x{:08X}",
condition_variable_addr, target);
- u32 processed = 0;
+ auto RetrieveWaitingThreads =
+ [](size_t core_index, std::vector<SharedPtr<Thread>>& waiting_threads, VAddr condvar_addr) {
+ const auto& scheduler = Core::System::GetInstance().Scheduler(core_index);
+ auto& thread_list = scheduler->GetThreadList();
+
+ for (auto& thread : thread_list) {
+ if (thread->condvar_wait_address == condvar_addr)
+ waiting_threads.push_back(thread);
+ }
+ };
+
+ // Retrieve a list of all threads that are waiting for this condition variable.
+ std::vector<SharedPtr<Thread>> waiting_threads;
+ RetrieveWaitingThreads(0, waiting_threads, condition_variable_addr);
+ RetrieveWaitingThreads(1, waiting_threads, condition_variable_addr);
+ RetrieveWaitingThreads(2, waiting_threads, condition_variable_addr);
+ RetrieveWaitingThreads(3, waiting_threads, condition_variable_addr);
+ // Sort them by priority, such that the highest priority ones come first.
+ std::sort(waiting_threads.begin(), waiting_threads.end(),
+ [](const SharedPtr<Thread>& lhs, const SharedPtr<Thread>& rhs) {
+ return lhs->current_priority < rhs->current_priority;
+ });
+
+ // Only process up to 'target' threads, unless 'target' is -1, in which case process
+ // them all.
+ size_t last = waiting_threads.size();
+ if (target != -1)
+ last = target;
+
+ // If there are no threads waiting on this condition variable, just exit
+ if (last > waiting_threads.size())
+ return RESULT_SUCCESS;
- auto signal_process_wide_key = [&](size_t core_index) {
- const auto& scheduler = Core::System::GetInstance().Scheduler(core_index);
- for (auto& thread : scheduler->GetThreadList()) {
- if (thread->condvar_wait_address != condition_variable_addr)
- continue;
+ for (size_t index = 0; index < last; ++index) {
+ auto& thread = waiting_threads[index];
- // Only process up to 'target' threads, unless 'target' is -1, in which case process
- // them all.
- if (target != -1 && processed >= target)
- break;
-
- // If the mutex is not yet acquired, acquire it.
- u32 mutex_val = Memory::Read32(thread->mutex_wait_address);
-
- if (mutex_val == 0) {
- // We were able to acquire the mutex, resume this thread.
- Memory::Write32(thread->mutex_wait_address, thread->wait_handle);
- ASSERT(thread->status == THREADSTATUS_WAIT_MUTEX);
- thread->ResumeFromWait();
-
- auto lock_owner = thread->lock_owner;
- if (lock_owner)
- lock_owner->RemoveMutexWaiter(thread);
-
- thread->lock_owner = nullptr;
- thread->mutex_wait_address = 0;
- thread->condvar_wait_address = 0;
- thread->wait_handle = 0;
- } else {
- // Couldn't acquire the mutex, block the thread.
- Handle owner_handle = static_cast<Handle>(mutex_val & Mutex::MutexOwnerMask);
- auto owner = g_handle_table.Get<Thread>(owner_handle);
- ASSERT(owner);
- ASSERT(thread->status != THREADSTATUS_RUNNING);
- thread->status = THREADSTATUS_WAIT_MUTEX;
- thread->wakeup_callback = nullptr;
-
- // Signal that the mutex now has a waiting thread.
- Memory::Write32(thread->mutex_wait_address, mutex_val | Mutex::MutexHasWaitersFlag);
-
- owner->AddMutexWaiter(thread);
-
- Core::System::GetInstance().CpuCore(thread->processor_id).PrepareReschedule();
- }
+ ASSERT(thread->condvar_wait_address == condition_variable_addr);
- ++processed;
- }
- };
+ // If the mutex is not yet acquired, acquire it.
+ u32 mutex_val = Memory::Read32(thread->mutex_wait_address);
+
+ if (mutex_val == 0) {
+ // We were able to acquire the mutex, resume this thread.
+ Memory::Write32(thread->mutex_wait_address, thread->wait_handle);
+ ASSERT(thread->status == THREADSTATUS_WAIT_MUTEX);
+ thread->ResumeFromWait();
- signal_process_wide_key(0);
- signal_process_wide_key(1);
- signal_process_wide_key(2);
- signal_process_wide_key(3);
+ auto lock_owner = thread->lock_owner;
+ if (lock_owner)
+ lock_owner->RemoveMutexWaiter(thread);
+
+ thread->lock_owner = nullptr;
+ thread->mutex_wait_address = 0;
+ thread->condvar_wait_address = 0;
+ thread->wait_handle = 0;
+ } else {
+ // Couldn't acquire the mutex, block the thread.
+ Handle owner_handle = static_cast<Handle>(mutex_val & Mutex::MutexOwnerMask);
+ auto owner = g_handle_table.Get<Thread>(owner_handle);
+ ASSERT(owner);
+ ASSERT(thread->status != THREADSTATUS_RUNNING);
+ thread->status = THREADSTATUS_WAIT_MUTEX;
+ thread->wakeup_callback = nullptr;
+
+ // Signal that the mutex now has a waiting thread.
+ Memory::Write32(thread->mutex_wait_address, mutex_val | Mutex::MutexHasWaitersFlag);
+
+ owner->AddMutexWaiter(thread);
+
+ Core::System::GetInstance().CpuCore(thread->processor_id).PrepareReschedule();
+ }
+ }
return RESULT_SUCCESS;
}